from abc import ABCMeta, abstractmethod


class BaseKD(object, metaclass=ABCMeta):
    @abstractmethod
    def tensorflow(self):
        pass

    @abstractmethod
    def torch(self):
        pass

    @abstractmethod
    def _check_config(self, config):
        pass


class ST(BaseKD):
    def __init__(self, alpha, T):
        self.config = {"alpha": alpha, "T": T}

    def _check_config(self, config):
        assert type(self.config["alpha"]) == float
        assert self.config["alpha"] <= 1.0 and self.config["alpha"] >= 0

        assert type(self.config["T"]) == float or type(self.config["T"]) == int

    def tensorflow(self):
        from .st_tensorflow import ST as TENSORFLOW_ST

        self._check_config(self.config)
        return TENSORFLOW_ST(self.config)

    def torch(self):
        from .st_torch import ST as TORCH_ST

        self._check_config(self.config)
        return TORCH_ST(self.config)



"""Ref
https://github.com/AberHu/Knowledge-Distillation-Zoo
https://github.com/sseung0703/Knowledge_distillation_via_TF2.0
"""


import torch
import torch.nn.functional as F
from .base_torch import BaseTorch


class ST(BaseTorch):
    def __init__(self, config):
        super(ST, self).__init__(config)

        self.teacher = None
        self.student = None
        self.optimizer = None

        self.alpha = config["alpha"]
        self.T = config["T"]

    def teach_step(self, x, y):
        x = x.to(self.device)
        y = y.to(self.device)
        self.optimizer.zero_grad()
        with torch.set_grad_enabled(False):
            t_logits = self.teacher(x)
        s_logits = self.student(x)
        self.logging_metrics(y, s_logits)
        loss = self.compute_loss(t_logits, s_logits, y)
        loss.backward()
        self.optimizer.step()
        return loss

    def compute_loss(self, t_logits, s_logits, labels):
        t_logits = t_logits.detach()
        return (1 - self.alpha) * self.cls_loss(
            labels, s_logits
        ) + self.alpha * self.st_loss(t_logits, s_logits)

    def cls_loss(self, labels, s_logits):
        criterion = torch.nn.CrossEntropyLoss()
        return criterion(s_logits, labels)

    def st_loss(self, t_logits, s_logits):
        assert t_logits.shape == s_logits.shape
        return (
            F.kl_div(
                F.log_softmax(s_logits / self.T, dim=1),
                F.softmax(t_logits / self.T, dim=1),
                size_average=False,
            )
            * (self.T ** 2)
            / s_logits.shape[0]
        )



def result_to_tqdm_template(result, training=True):
    template = ""
    for k in result.keys():
        display_name = k
        if not training:
            display_name = "val_{}".format(display_name)
        template += "{}: {:.4f} -".format(display_name, result[k])
    return template[:-1]



"""Ref
https://arxiv.org/pdf/1503.02531.pdf
https://github.com/sseung0703/Knowledge_distillation_via_TF2.0
"""

import tensorflow as tf
from .base_tensorflow import BaseTensorflow


class ST(BaseTensorflow):
    def __init__(self, config):
        super(ST, self).__init__(config)

        self.teacher = None
        self.student = None
        self.optimizer = None

        self.alpha = config["alpha"]
        self.T = config["T"]

    def teach_step(self, x, y):
        with tf.GradientTape() as tape:
            t_logits = self.teacher(x, training=False)
            s_logits = self.student(x, training=True)
            loss = self.compute_loss(t_logits, s_logits, y)
        grad = tape.gradient(loss, self.student.trainable_variables)
        self.optimizer.apply_gradients(zip(grad, self.student.trainable_variables))
        self.logging_metrics(y, s_logits)
        return loss

    @tf.function
    def compute_loss(self, t_logits, s_logits, labels):
        cls_loss = self.cls_loss(labels, s_logits)
        st_loss = self.st_loss(t_logits, s_logits)
        return (1 - self.alpha) * cls_loss + self.alpha * st_loss

    @tf.function
    def cls_loss(self, labels, s_logits):
        criterion = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)
        return criterion(labels, s_logits)

    @tf.function
    def st_loss(self, t_logits, s_logits):
        assert t_logits.shape == s_logits.shape
        return tf.reduce_mean(
            tf.reduce_sum(
                tf.nn.softmax(t_logits / self.T)
                * (
                    tf.nn.log_softmax(t_logits / self.T)
                    - tf.nn.log_softmax(s_logits / self.T)
                ),
                1,
            )
        )



from abc import ABCMeta, abstractmethod

import torch
import ignite


class BaseTorch(object, metaclass=ABCMeta):
    def __init__(self, config):
        self.metrics = {"accuracy": ignite.metrics.Accuracy()}
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    def set_model(self, teacher, student, optimizer):
        self.teacher = teacher.to(self.device)
        self.student = student.to(self.device)
        self.optimizer = optimizer

    @abstractmethod
    def teach_step(self, x, y):
        pass

    def logging_metrics(self, labels, logits):
        for name in self.metrics.keys():
            self.metrics[name].update([logits, labels])

    def reset_metrics(self):
        for name in self.metrics.keys():
            self.metrics[name].reset()

    def set_metrics(self, metrics):
        for name in metrics.keys():
            assert isinstance(metrics.keys()[name], ignite.metrics.Metric)
            self.metrics[name] = metrics

    def get_metrics(self):
        result = {}
        for name in self.metrics.keys():
            result[name] = self.metrics[name].compute()
        return result

    def test(self, dataset):
        self.reset_metrics()
        self.student.eval()
        with torch.set_grad_enabled(False):
            for x, y in dataset:
                x, y = x.to(device), y.to(device)
                logits = self.student(x)
                self.logging_metrics(y, logits)
        self.student.train()
        result = self.get_metrics()
        self.reset_metrics()
        return result



from setuptools import setup, find_packages

setup(
    name="aquvitae",
    version="0.1.2",
    description="The easiest Knowledge Distillation library for Light Weight DeepLearning",
    author="marload",
    author_email="rladhkstn8@gmail.com",
    url="https://github.com/aquvitae/aquvitae",
    download_url="https://github.com/aquvitae/aquvitae/archive/v0.1.0.tar.gz",
    license="MIT",
    install_requires=[],
    setup_requires=["tqdm>=4.46.1", "pytorch-ignite>=0.4.0"],
    packages=find_packages(exclude=[]),
    keywords=["tensorflow", "pytorch", "pytorch-ignite", "tqdm"],
    python_requires=">=3",
    package_data={},
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)



import os
from tqdm import trange

import torch
import tensorflow as tf


from .algo import BaseKD
from .utils import result_to_tqdm_template


def dist(
    teacher,
    student,
    algo,
    optimizer,
    train_ds,
    test_ds,
    iterations,
    metrics={},
    test_freq=200,
):
    info = _check_dist(teacher, student, algo, optimizer, train_ds, test_ds)

    if info["framework"] == "torch":
        kd = algo.torch()
        kd.set_metrics(metrics)
        return _torch_dist(
            teacher.eval(),
            student,
            kd,
            optimizer,
            train_ds,
            test_ds,
            iterations,
            test_freq,
        )
    elif info["framework"] == "tensorflow":
        kd = algo.tensorflow()
        kd.set_metrics(metrics)
        return _tensorflow_dist(
            teacher, student, kd, optimizer, train_ds, test_ds, iterations, test_freq,
        )
    else:
        raise NotImplementedError


def _check_dist(teacher, student, algo, optimizer, train_ds, test_ds):
    assert isinstance(algo, BaseKD)

    info = {"framework": None}
    if isinstance(teacher, torch.nn.Module) and isinstance(student, torch.nn.Module):
        assert isinstance(optimizer, torch.optim.Optimizer)
        assert isinstance(train_ds, torch.utils.data.DataLoader)
        assert isinstance(test_ds, torch.utils.data.DataLoader)
        info["framework"] = "torch"
    elif isinstance(teacher, tf.keras.Model) and isinstance(student, tf.keras.Model):
        assert isinstance(optimizer, tf.keras.optimizers.Optimizer)
        assert isinstance(train_ds, tf.data.Dataset)
        assert isinstance(test_ds, tf.data.Dataset)
        info["framework"] = "tensorflow"

    return info


def _init_dist(teacher, student, algo, optimizer, iterations):
    algo.set_model(teacher, student, optimizer)
    bar_format = "{desc} - {n_fmt}/{total_fmt} [{bar:30}] ELP: {elapsed}{postfix}"
    process_log = trange(iterations, desc="Training", position=0, bar_format=bar_format)
    return algo, process_log


def _tensorflow_dist(
    teacher, student, algo, optimizer, train_ds, test_ds, iterations, test_freq
):
    algo, process_log = _init_dist(teacher, student, algo, optimizer, iterations)

    train_tmp = ""
    test_tmp = ""

    train_ds = train_ds.repeat()
    for idx, (x, y) in enumerate(train_ds):
        process_log.update(1)
        loss = algo.teach_step(x, y)
        result = algo.get_metrics()
        train_tmp = result_to_tqdm_template(result)

        if idx % test_freq == 0 and idx != 0:
            process_log.set_description_str("Testing ")
            algo.reset_metrics()
            result = algo.test(test_ds)
            test_tmp = result_to_tqdm_template(result, training=False)
            process_log.set_description_str("Training")
        postfix = train_tmp + "- " + test_tmp
        process_log.set_postfix_str(postfix)

    return student


def _torch_dist(
    teacher, student, algo, optimizer, train_ds, test_ds, iterations, test_freq
):
    algo, iter_log, train_metric_log, test_metric_log = _init_dist(
        teacher, student, algo, optimizer, iterations
    )

    train_ds = iter(train_ds)

    for idx in range(iterations):
        try:
            x, y = train_ds.next()
        except StopIteration:
            train_ds = iter(train_ds)
            x, y = train_ds.next()
        loss = algo.teach_step(x, y)
        result = algo.get_metrics()
        train_metric_tmp = result_to_tqdm_template(result)
        train_metric_log.set_description_str("TRAIN:\t" + train_metric_tmp)
        iter_log.update(1)

        if idx % test_freq == 0 and idx != 0:
            algo.reset_metrics()
            result = algo.test(test_ds)
            test_metric_tmp = result_to_tqdm_template(result)
            test_metric_log.set_description_str("TEST:\t" + test_metric_tmp)

    return student



from abc import ABCMeta, abstractmethod

import tensorflow as tf


class BaseTensorflow(object, metaclass=ABCMeta):
    def __init__(self, config):
        self.metrics = {"accuracy": tf.keras.metrics.SparseCategoricalAccuracy()}

    def set_model(self, teacher, student, optimizer):
        self.teacher = teacher
        self.student = student
        self.optimizer = optimizer

    @abstractmethod
    def teach_step(self, x, y):
        pass

    def logging_metrics(self, labels, logits):
        for name in self.metrics.keys():
            self.metrics[name].update_state(labels, logits)

    def reset_metrics(self):
        for name in self.metrics.keys():
            self.metrics[name].reset_states()

    def set_metrics(self, metrics):
        for name in metrics.keys():
            assert isinstance(metrics.keys()[name], tf.keras.metrics.Metric)
            self.metrics[name] = metrics

    def get_metrics(self):
        result = {}
        for name in self.metrics.keys():
            result[name] = self.metrics[name].result()
        return result

    def test(self, dataset):
        self.reset_metrics()
        for x, y in dataset:
            logits = self.student(x)
            self.logging_metrics(y, logits)

        result = self.get_metrics()
        self.reset_metrics()
        return result
